"""The docker tool provides ``build``, ``push`` and ``publish`` commands.

The `build` tasks build the dockerfile and tags the resulting image with the
project's name.
"""
import argparse
import logging
import os

from functools import reduce

from menhir.tool import DEPLOY, Tool
from menhir.tool_utils import (
    FAIL,
    NOTHING_TO_DO,
    OK,
    argv_to_dict,
    call,
    changed_state,
    has_self_or_dependent_changes,
    tool_env,
    working_dir,
)
from menhir.utils import multi, method

log = logging.getLogger(__name__)


def tool():
    return Docker()


class Docker(Tool):

    def name(arg):
        return "docker"

    def dir_info(tool, path, info):
        from os.path import exists, join
        path = join(path, 'Dockerfile')
        has_dockerfile = exists(path)
        return {
            'project_recognised': has_dockerfile,
            'can_run': has_dockerfile,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build phase."""
        from os.path import exists, join
        from menhir.tool_utils import run_if

        phase_name = args.phase

        dockerfile = join(path, 'Dockerfile')

        if not exists(dockerfile):
            log.debug(
                'No Dockerfile %(dockerfile)s',
                {'dockerfile': dockerfile})
            return NOTHING_TO_DO

        with working_dir(path):
            changed = changed_state(info)
            run_flag = has_self_or_dependent_changes(changed)

            if phase_name in {'build', 'push', 'publish'}:
                with run_if(run_flag, phase_name, path) as flag:
                    if flag:
                        return task(phase_name, path, info, args)
                    return OK
            else:
                return NOTHING_TO_DO

    def build_jobs(tool, info):
        return [DEPLOY]


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Commands to build and push docker images.",
        **kwargs
    )
    parsers = parser.add_subparsers(help="Docker commands", dest='phase')
    p = parsers.add_parser(
        'build',
        help='Build a docker image from a Dockerfile'
    )
    p.add_argument(
        '--build-arg', dest='build_args', action='append',
        help="Specify a build argument for the docker build")
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'push',
        help='Push a docker image to a remote repository'
    )
    p.add_argument(
        '--arg', dest='args', action='append',
        help="Specify an argument for the encryption settings")
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'publish',
        help='Build and push a docker image to a remote repository'
    )
    p.add_argument(
        '--arg', dest='args', action='append',
        help="Specify an argument for the encryption settings")
    p.add_argument(
        '--build-arg', dest='build_args', action='append',
        help="Specify a build argument for the docker build")
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    return parser


@multi
def task(phase_name, path, info, args):
    return phase_name


@method(task)
def task_default(phase_name, path, info, args):
    return NOTHING_TO_DO


@method(task)
def task_publish(phase_name, path, info, args):
    res = task('build', path, info, args)
    if res != OK:
        return res
    return task('push', path, info, args)


@method(task, 'build')
def docker_build(phase_name, path, info, args):
    log.info('Running docker-build in %s', path)

    project_name = info['project-name']

    env = tool_env()

    env_names = [
        v for v in info.get('docker', {}).get('env', [])
        if v in os.environ
    ]

    for v in env_names:
        env[v] = os.environ[v]

    env_args = reduce(
        lambda v, a: v + ['--build-arg', '{}={}'.format(a, os.environ[a])],
        env_names,
        []
    )

    build_args = reduce(
        lambda v, a: v + ['--build-arg', a],
        args.build_args or [],
        [])

    # docker build -t "${MENHIR_TAG}" "$@" .
    argv = (
        ['docker', 'build', '-t', project_name] + build_args + env_args + ['.']
    )
    print(argv)
    return call(argv, env=env)


@method(task, 'push')
def docker_push(phase_name, path, info, args):
    from menhir.project import branch, image
    from menhir.tool_utils import call, package_script, slugify
    log.info('Running docker-push in %s', path)

    arg_names = info.get('docker', {}).get('args', [])
    arg_values = args.args or []

    arg_dict = argv_to_dict(arg_names, arg_values)
    if arg_dict is None:
        log.error('Expected --arg for each of %s', arg_names)
        return FAIL

    project_name = info['project-name']
    current_branch = branch()
    tag = project_name
    sha_tag = image(info, path)
    sha_tag = sha_tag % arg_dict
    if not sha_tag:
        log.error('No remote repository configured to push to.')
        return FAIL
    branch_tag = "%s:%s" % (
        sha_tag.split(':')[0],
        slugify(current_branch, length=40),
    )
    branch_tag = branch_tag % arg_dict

    env = tool_env()
    env['MENHIR_TAG'] = tag
    env['MENHIR_BRANCH_TAG'] = branch_tag
    env['MENHIR_SHA_TAG'] = sha_tag

    with package_script("/tools/docker/docker-push.sh") as f:
        return call([f.name], env=env,)



"""The codecov tool uplods coverage data to ``codecov.io``."""
import argparse
import logging

from os.path import exists, join

from menhir.tool import TEST, Tool
from menhir.tool_utils import (
    NOTHING_TO_DO,
    OK,
    changed_state,
    has_self_or_dependent_changes,
    tool_env,
    working_dir,
)

log = logging.getLogger(__name__)


def tool():
    return Codecov()


class Codecov(Tool):
    def name(arg):
        return "codecov"

    def dir_info(tool, path, info):
        setup_path = join(path, 'setup.cfg')
        path = join(path, '.coverage')
        return {
            'project_recognised': exists(setup_path),
            'can_run': exists(path),
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build phase."""
        import re
        from os import getenv
        from os.path import exists, join
        from menhir.tool_utils import call, package_script

        if not getenv('CODECOV_TOKEN'):
            return NOTHING_TO_DO

        log.info('Try Running codecov in %s', path)

        with working_dir(path):

            changed = changed_state(info)

            if has_self_or_dependent_changes(changed):
                log.info('Running codecov in %s', path)

                if not exists(join(path, '.coverage')):
                    log.debug('No .coverage in %s', path)
                    return OK

                env = tool_env()
                env['MENHIR_PROJECT'] = info['project-name']
                env['MENHIR_CODECOV_FLAGS'] \
                    = re.sub(r'\W', "_", info['project-name'])
                env['CODECOV_TOKEN'] = getenv('CODECOV_TOKEN')

                with package_script("/tools/codecov/upload.sh") as f:
                    return call([f.name], env=env,)
            else:
                log.info('not running codecov in %s', path)
                return OK

    def build_jobs(tool, info):
        return [TEST]


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Push coverage metrics to codecov.io",
        **kwargs
    )
    parser.add_argument('remainder', nargs=argparse.REMAINDER)
    return parser



"""Terraform tool.

Provides ``plan``, ``apply`` and ``destroy`` tasks for terrform, managing
remote-state.

Configuration is under the ``terraform`` key.

The ``targets`` key contains a dict for each target directory under
``source-dir`` (defaults to ``infra``).  Each infra directory is
configured as a dict under the directory's name.

Arguments
---------

A directory can be configures with ``args``, a list containing the names
of values it expects to be passed on the command line.


Vars passed to Terraform
------------------------

The ``values`` key configures a list of names for values to be passed
to terraform.

``values`` can contain names from ``args`` and the following:

project        the name of the project
branch         the current branch name
branch-slug    the current branch name, sanitised
sha            the current commit sha as a hex string
sha-mod-1024   the current as a decimal modulo 1024

The ``values`` list can also contain items of the form ``name=value``,
which will pass the var ``name`` to terraform, with the value
calculated according to ``value`` from the list above.

Remote Config
-------------

The remote state is configured independently of the target direcotry
using the `remote-config` key.  The value contains a `backend` string
specifying the terraform backend, and a `backend-config` dict
specifying the backend specific configuration.

The `remote-config` key is also supported at the target directory
level.  For consistency, usage at this level is not recommended.

The only supported backend at present is ``s3``.  It expects
``bucket``, ``key`` and ``region`` values.  The ``key`` configuration
value is a format string, with can use any of the ``values`` and
``args`` configuration values.  It can also use ``target`` to refer to
the target name.

"""
import argparse
import logging
import subprocess

from functools import reduce

from menhir.tool import Tool
from menhir.tool_utils import FAIL, NOTHING_TO_DO, OK, tool_env, working_dir
from menhir.utils import multi, method

log = logging.getLogger(__name__)


def tool():
    return Terraform()


class Terraform(Tool):

    def dir_info(tool, path, info):
        from os.path import isdir, join
        path = join(
            path,
            info.get('terraform', {}).get('source-dir', 'infra')
        )
        has_terraform = isdir(path)
        return {
            'project_recognised': has_terraform,
            'can_run': has_terraform,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        """Execute a build phase."""
        config = info.get('terraform', {})
        return task(path, info, config, args)

    def build_jobs(tool, info):
        return []


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Tool for terraform."
    )
    parsers = parser.add_subparsers(
        help="Terraform commands",
        dest='phase'
    )

    p = parsers.add_parser(
        'apply',
        help='Apply a terraform infrastructure directory'
    )
    p.add_argument('directory', nargs=1)
    p.add_argument('args', nargs="*")

    p = parsers.add_parser(
        'plan',
        help='Plan a terraform infrastructure directory'
    )
    p.add_argument('directory', nargs=1)
    p.add_argument('args', nargs="*")

    p = parsers.add_parser(
        'destroy',
        help='Destroy a terraform infrastructure directory'
    )
    p.add_argument('directory', nargs=1)
    p.add_argument('args', nargs="*")

    p = parsers.add_parser(
        'install',
        help="Download and install terraform"
    )

    os, arch = platform()
    p.add_argument('ver', nargs=1, help='version')
    p.add_argument(
        '--os',
        default=os,
        help='Operating system',
        choices=['darwin', 'linux', 'freebsd', 'openbsd', 'solaris'],
    )
    p.add_argument(
        '--arch',
        default=arch,
        help='Architecture',
        choices=['amd64', '386', 'arm'],
    )
    p.add_argument(
        '--target',
        default='/usr/local/bin',
        help='target directory')
    p.add_argument(
        '--source',
        default='https://releases.hashicorp.com/terraform',
        help='Distribution source')
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    return parser


@multi
def task(path, info, config, args):
    """Execute a task."""
    return args.phase


@method(task)
def task_action(path, info, config, args):
    return action(args.phase, path, info, config, args)


def action(action, path, info, config, args):
    """Invoke a terraform action on the given path."""
    from os import getenv
    from os.path import exists, join
    from menhir.tool_config import aliased_value_array
    from menhir.tool_utils import call
    from menhir.utils import deep_merge

    log.debug('terraform config: %(config)s', dict(config=config))
    infra_dir = join(path, config.get('source-dir', 'infra'))

    if not exists(infra_dir):
        log.debug('No infra directory in %(path)s', {'path': path})
        return NOTHING_TO_DO

    directory = args.directory[0]
    run_flag = (
        'changed' not in info or
        info['changed'].get('self') or
        info['changed'].get('dependents')
    )
    project_dir = join(infra_dir, args.directory[0])
    project_dir_exists = exists(project_dir)

    run = project_dir_exists and (run_flag or action == 'destroy')
    log_run(run, args.phase, directory, path)

    if not run:
        return NOTHING_TO_DO

    dconfig = config.get('targets', {}).get(directory, {})
    dargs = dconfig.get('args', [])
    dvalues = dconfig.get('values', [])
    remote_config = config.get('remote-config')
    dremote_config = dconfig.get('remote-config', {})
    remote_config = deep_merge(remote_config or {}, dremote_config)

    if len(args.args or []) < len(dargs):
        log.error(
            'Not enough arguments to terraform apply %(directory)s: '
            'expected: %(expected)s, got: %(got)s',
            {
                'directory': directory,
                'expected': dargs,
                'got': args.args,
            }
        )
        return FAIL

    arg_dict = dict(zip(dargs, args.args))

    def format_value(x):
        return '%s="%s"' % x

    values_list = aliased_value_array(dvalues, info, path, arg_dict)
    values = list(map(format_value, values_list))

    values_dict = {k: v for k, v in values_list}
    values_and_args = {
        'target': directory,
        'project': info['project-name'],
    }
    values_and_args.update(arg_dict)
    values_and_args.update(values_dict)

    values_and_args = {
        k: v % values_and_args if isinstance(v, str) else v
        for k, v in values_and_args.items()
    }

    def extend_var(x, y):
        x.extend(['-var', y])
        return x

    var_args = reduce(extend_var, values, [],)

    project_name = info['project-name']

    env = tool_env()

    tf_log = getenv('TF_LOG')
    if tf_log:
        env['TF_LOG'] = tf_log

    with working_dir(project_dir):
        if remote_state(remote_config, project_name, values_and_args):
            return FAIL
        log.info("terraform get %s", env)
        res = call(["terraform", "get"], env=env,)
        if res != OK:
            return res
        log.info("terraform %s %s", action, var_args)
        return call(["terraform", action] + var_args, env=env,)


def log_run(cond, task, directory, path):
    txt = None
    if cond:
        txt = 'Run %(task)s in %(path)s'
    else:
        txt = 'Not run %(task)s in %(path)s'

    log.info(txt, {'task': task, 'directory': directory, 'path': path})


def remote_state(remote_config, project, values):
    import shutil

    tf_version = terraform_version()
    log.debug('terraform version: %s', tf_version)

    use_remote_config = tf_version < "0.9.0"
    backend = remote_config.get('backend')
    backend_config = remote_config.get('backend-config')

    if not backend:
        log.warning('remote-state configured, but no backend specified')
        return None

    if not backend_config:
        log.warning(
            'remote-state configured, but no backend_config specified'
        )
        return None

    shutil.rmtree('.terraform', ignore_errors=True)
    return remote_state_backend(
        backend, backend_config, project, values, use_remote_config
    )


def terraform_version():
    proc = subprocess.Popen(["terraform", "--version"], stdout=subprocess.PIPE)
    lines = proc.stdout.readlines()
    if lines:
        words = lines[0].split()
        version = words[1]
        return version[1:].decode('utf-8')  # drop the leading 'v'


@multi
def remote_state_backend(
        backend, backend_config, project, values, use_remote_config
):
    """Configure a remote backend."""
    return backend


@method(remote_state_backend, 's3')
def remote_state_s3(
        backend, backend_config, project, values, use_remote_config
):
    from os import getenv
    from os.path import join
    from subprocess import call

    bucket = backend_config.get('bucket')
    if not bucket:
        log.warning('terraform remote-state s3: no bucket specified')
        return None

    env = tool_env()

    tf_log = getenv('TF_LOG')
    if tf_log:
        env['TF_LOG'] = tf_log

    if use_remote_config:
        key = backend_config['key']
        key = key % values
        key = join(key, "terraform.tfstate")
        region = backend_config.get('region', "us-east-1")

        log.info(
            'State from s3://%(bucket)s/%(key)s in %(region)s',
            {
                'bucket': bucket,
                'key': key,
                'region': region,
            }
        )

        cmd = [
                "terraform", "remote", "config",
                "-backend=s3",
                "-backend-config=bucket=%s" % bucket,
                "-backend-config=key=%s" % key,
                "-backend-config=region=%s" % region,
            ]
        log.debug('%s %s', cmd, env)

        return call(cmd, env=env)
    else:

        def extend_var(x, y):
            x.extend(['-backend-config=%s=%s' % y % values])
            return x

        bc = reduce(extend_var, backend_config.items(), [],)

        cmd = [
            "terraform", "init",
        ] + bc

        log.debug('%s %s', cmd, env)
        return call(cmd, env=env)


# Guard against installation from multiple projects
installed = False


@method(task, 'install')
def task_install(path, info, config, args):
    import shutil
    import stat
    import zipfile

    from os import chmod
    from os.path import join
    from tempfile import TemporaryFile
    from tqdm import tqdm
    import requests

    global installed

    if installed:
        return NOTHING_TO_DO

    version = args.ver[0]
    os = args.os
    arch = args.arch
    target = args.target

    file_name = 'terraform_%s_%s_%s.zip' % (version, os, arch)
    url = "%s/%s/%s" % (
        args.source,
        version,
        file_name
    )
    target_file = join(target, 'terraform')

    print(url)
    with TemporaryFile() as f:
        response = requests.get(url, stream=True)
        for data in tqdm(response.iter_content(chunk_size=1024*1024)):
            f.write(data)

        with zipfile.ZipFile(f) as z:
            with z.open("terraform") as zf, open(target_file, 'wb') as f:
                shutil.copyfileobj(zf, f)

    chmod(target_file, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)

    installed = True
    return OK


def platform():
    import platform
    arch = {
        'x86_64': 'amd64',
    }

    system = platform.system().lower()
    machine = platform.machine()

    return system, arch.get(machine, machine)



"""Tool that writes the project state to a file."""

import argparse
import logging

from menhir.fileutils import yaml_dumper
from menhir.tool import Tool
from menhir.tool_utils import OK, NOTHING_TO_DO, working_dir

log = logging.getLogger(__name__)


def tool():
    return WriteState()


class WriteState(Tool):

    def dir_info(tool, path, info):
        return {
            'project_recognised': False,
            'can_run': True,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        """Arg parser for the tool options and arguments."""
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        """Execute the write_state tool."""
        with working_dir(path):
            log.info('Writing state in %s', path)
            if 'changed' in info:
                with open('menhir-state.yaml', 'w+') as f:
                    yaml = yaml_dumper()
                    yaml.dump({'changed': info['changed']}, f)
                    return OK

        return NOTHING_TO_DO


def parser(**kwargs):
    """Arg parser for the tool options and arguments."""
    parser = argparse.ArgumentParser(
        description="Write state.",
        **kwargs
    )
    return parser



"""The gemfury tool provides ``publish`` command.

The ``publish`` task takes a file and pushes it to gemfury.

Configuration is under the ``gemfury`` key.

The ``org`` key contains the user or organisation name to publish artifacts to
on GemFury.

"""
import argparse
import logging

import requests

from menhir.tool import Tool
from menhir.tool_utils import (
    OK,
    FAIL,
    NOTHING_TO_DO,
)
from menhir.utils import multi, method

log = logging.getLogger(__name__)


def tool():
    return GemFury()


class GemFury(Tool):

    def name(arg):
        return "gemfury"

    def dir_info(tool, path, info):
        from os.path import exists, join
        path = join(path, '.gemfury')
        recognised = info.get('gemfury') or exists(path)
        return {
            'project_recognised': recognised,
            'can_run': recognised,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build phase."""
        return NOTHING_TO_DO


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Commands to interact with gemfury.",
        **kwargs
    )
    parsers = parser.add_subparsers(help="GemFury commands", dest='phase')
    p = parsers.add_parser(
        'publish',
        help='Publish a release to GemFury'
    )
    p.add_argument(
        '--org',
        help="Specify the GemFury user or organisation name to publish to")
    p.add_argument(
        '--token',
        help="Specify the GemFury token to use for publishing")
    p.add_argument(
        '--package',
        help="path to package to publish")
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    return parser


@multi
def task(phase_name, path, info, args):
    return phase_name


@method(task)
def task_default(phase_name, path, info, args):
    return NOTHING_TO_DO


@method(task, 'publish')
def gemfury_build(phase_name, path, info, args):
    log.info('Running gemfury publish in %s', path)

    gemfury = info.get('gemfury', {})
    token = args.token or gemfury.get('token')
    org = args.org or gemfury.get('org')
    package = args.package or gemfury.get('package')

    if not token:
        log.error('No gemfury token specified')
        return FAIL

    url = "https://{}@push.fury.io/{}/".format(
        token,
        org,
    )
    with open(package) as f:
        res = requests.post(url, files=dict(package=f))

    res.raise_for_status()

    return OK



"""Utilities for use in tools."""

import logging
import os.path
import re
import sys
from contextlib import contextmanager

from menhir.fileutils import load_yaml
from menhir.tool import load_tool

log = logging.getLogger(__name__)

OK = {
    'status': 'ok'
}

FAIL = {
    'status': 'fail'
}

NOTHING_TO_DO = {
    'status': 'nothing_to_do'
}

NON_WORD_PATTERN = re.compile(r'\W')


@contextmanager
def package_script(resource_path, resource_package="menhir"):
    """Execute a block of code with the given script from the package.

    Yields a file like object that is the script written onto the filesystem.
    """
    import tempfile
    import pkg_resources
    import stat
    from os import chmod, remove

    script = pkg_resources.resource_string(resource_package, resource_path)
    fname = None
    try:
        with tempfile.NamedTemporaryFile("wb", delete=False) as f:
            fname = f.name
            f.write(script)
        chmod(fname, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR)
        yield f
    finally:
        remove(fname)


@contextmanager
def working_dir(path):
    """Execute a block of code within the given working dir."""
    import os
    dirname = os.getcwd()
    log.debug('Change working dir from %s to %s', dirname, path)
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(dirname)


ENV_TO_REMOVE = [
    'PWD',
    'PYENV_DIR',
    'PYENV_HOOK_PATH',
    'PYENV_SHELL',
    'PYENV_VERSION',
    'PYENV_VIRTUALENV_INIT',
]


def tool_env():
    """Return the default tool environment dict."""
    import os
    import os.path

    default_pyenv_root = os.path.join(os.getenv("HOME"), ".pyenv")
    env = {
        "PYENV_ROOT": default_pyenv_root,
    }

    env.update(os.environ)

    for k in ENV_TO_REMOVE:
        env.pop(k, None)
    return env


def call(cmd, *args, **kwargs):
    """Call a subprocess, returning a menhir result."""
    import subprocess
    res = subprocess.call(cmd, *args, **kwargs)
    if res:
        return FAIL
    return OK


def slugify(s, length=None, replace=NON_WORD_PATTERN):
    s = re.sub(replace, "_", s)
    if length:
        s = s[:length]
    return s


@contextmanager
def run_if(cond, phase_name, path):
    if cond:
        log.info(
            'Running %(phase)s in %(path)s',
            {'phase': phase_name, 'path': path}
        )
        yield True
    else:
        yield False
        log.info(
            'Not running %(phase)s in %(path)s',
            {'phase': phase_name, 'path': path}
        )


def call_tool_chain(path, info, remainder):
    result = OK
    while remainder:
        tool_name = remainder.pop(0)
        tool = load_tool(tool_name)

        parser = tool.arg_parser(add_help=False, prog=tool_name)
        tool_args, unknown = parser.parse_known_args(
            remainder,
            # namespace=copy(args)
        )
        log.debug('tool_args %s: %s', tool_name, tool_args)
        if unknown:
            log.debug('unknown argument: %s', unknown)
            parser.print_help()
            sys.exit(0 if remainder == ['-h'] else 1)

        result = tool.execute_tool(path, info, tool_args)

        remainder = getattr(tool_args, 'remainder', None)
        log.debug('remainder after %s: %s', tool_name, remainder)

        if result['status'] != 'ok':
            return result

    return result


def argv_to_dict(arg_names, arg_values):
    """Convert vectors of arg names and arg values to a dict.

    returns `None, Fail` if the vector lengths do not match.
    """
    if len(arg_names) != len(arg_values):
        print(len(arg_names), len(arg_values))
        return None
    else:
        return {k: v for (k, v) in zip(arg_names, arg_values)}


def has_self_or_dependent_changes(changed):
    return (
        changed is None or
        changed.get('self') or
        changed.get('dependents')
    )


def load_menhir_state():
    p = 'menhir-state.yaml'
    if os.path.exists(p):
        return load_yaml(p)


def changed_state(info):
    changed = info.get('changed')
    if changed is None:
        changed = load_menhir_state()
    return changed



"""The setup_py tool provides functionality based on ``setup.py``.

The setup.py tool contributes iner-project dependencies by reading
relative ``file://...`` links in ``requirements.txt`` files.

The setup_py tool provides the ``sdist`` and ``lambda-package``
tasks.

lambda-package
--------------

The ``lambda-package`` task builds a ``package.zip`` file for upload
to AWS Lambda.  The tool is only run if there is a .lambda-package
file in the project directory.

"""
from __future__ import print_function  # NOQA

import argparse
import logging
import os
import re
import subprocess
import sys

from menhir.tool import DEPENDENCIES, PACKAGE, TEST, Tool
from menhir.tool_utils import (
    FAIL,
    NOTHING_TO_DO,
    OK,
    changed_state,
    has_self_or_dependent_changes,
    package_script,
    tool_env,
    working_dir
)
from menhir.utils import method, multi

DEP_PATTERN = re.compile('file:(.*)#egg=.*')

log = logging.getLogger(__name__)


def tool():
    return SetupPy()


class SetupPy(Tool):

    def name(arg):
        return "setup_py"

    def dir_info(tool, path, info):
        from os.path import exists, join
        has_setup_py = exists(join(path, 'setup.py'))
        return {
            'project_recognised': has_setup_py,
            'can_run': has_setup_py,
        }

    def dependencies(tool, path):
        from os.path import join, normpath
        log.debug('setup.py dependencies in %s', path)
        requirements_path = join(path, 'requirements.txt')
        if not os.path.isfile(requirements_path):
            return []

        try:
            requirements, dependency_links = read_requirements(
                requirements_path
            )
            return [
                normpath(join(path, DEP_PATTERN.match(l).group(1)))
                for l in dependency_links
                if l.startswith('file:')
            ]
        except Exception:
            print('Failed to infer dependencies in for setup_py %s' % path,
                  file=sys.stderr)
            raise

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        """Execute a build phase."""
        with working_dir(path):
            changed = changed_state(info)
            if has_self_or_dependent_changes(changed):
                return task(path, info, args)
        return NOTHING_TO_DO

    def build_jobs(tool, info):
        return [DEPENDENCIES, TEST, PACKAGE]


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Invoke helpers that use setup.py",
        **kwargs
    )
    parsers = parser.add_subparsers(help="Tool phases", dest='phase')

    p = parsers.add_parser(
        'sdist',
        help='Run dist on project, and on local project requirements'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'lambda-package',
        help='Build a package.zip file for the setup.py project'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'test',
        help='Run tests for the setup.py project'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'install',
        help='Install the setup.py project'
    )
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    return parser


def read_requirements(path):
    with open(path, 'r') as f:
        return process_requirements(f)


def process_requirements(f):
    requirements = []
    dependency_links = []
    for line in f.read().split():
        if '#egg=' in line:
            package_name = line.rsplit('#egg=', 1).pop()
            requirements.append(package_name)
            dependency_links.append(line)
        else:
            requirements.append(line)
    return requirements, dependency_links


@multi
def task(path, info, args):
    log.info('Running %s in %s', args.phase, path)
    return args.phase


@method(task)
def default_task(path, info, args):
    """Unknown task, does nothing."""
    return NOTHING_TO_DO


@method(task, 'sdist')
def sdist(pdir, info, args):
    """Tar python project at path, including relative requirements."""
    import os.path as path
    import shutil
    import subprocess
    from os import mkdir, walk

    res = subprocess.call(["python", "setup.py", "sdist"], env=tool_env())
    if res:
        return FAIL

    dist_reqs = 'dist-requirements'
    if not path.exists(dist_reqs):
        mkdir(dist_reqs)

    reqs = "requirements.txt"
    if path.exists(reqs):
        with open(reqs) as f:
            for line in f.read().splitlines():
                match = DEP_PATTERN.match(line)
                if match:
                    dir = match.group(1)
                    with working_dir(dir):
                        log.debug('Running sdist for %s', dir)
                        res = subprocess.call(
                            ["python", "setup.py", "sdist"],
                            env=tool_env(),
                        )
                        if res:
                            return False
                    for root, dirs, files in walk(path.join(dir, 'dist')):
                        for f in files:
                            shutil.copyfile(
                                path.join(root, f),
                                path.join(dist_reqs, f))

    log.debug('Completed sdist in %s', pdir)
    return OK


@method(task, 'lambda-package')
def lambda_package(path, info, args):
    """Zip python project at path, including relative requirements.

    Only packages projects that contain a ``.lambda-package`` file.
    """
    import subprocess
    from os.path import exists
    from menhir.tool_utils import package_script

    marker = '.lambda-package'
    serverless = 'serverless.yml'
    if not (exists(marker) or exists(serverless)):
        print('Not a lambda project (no %s)' % marker)
        return NOTHING_TO_DO

    with package_script("/tools/setup_py/lambda-package.sh") as f:
        res = subprocess.call([f.name], env=tool_env())
        if res:
            return FAIL
    return OK


@method(task, 'safety')
def safety(path, info, args):
    """Run safety vulnerability checks."""
    with package_script("/tools/setup_py/safety.sh") as f:
        res = subprocess.call([f.name], env=tool_env())
        if res:
            return FAIL
    return OK


@method(task, 'test')
def test(path, info, args):
    """Run tests."""
    with package_script("/tools/setup_py/wrapper.sh") as f:
        res = subprocess.call(
            [f.name, 'python', 'setup.py', 'test'],
            env=tool_env(),
        )
    if res:
        return FAIL
    return OK


@method(task, 'install')
def install(path, info, args):
    """Run install."""
    with package_script("/tools/setup_py/wrapper.sh") as f:
        res = subprocess.call([f.name, 'pip', 'install', '.'], env=tool_env())
    if res:
        return FAIL
    return OK



"""The build tool invokes build phases according to a configuration."""
import argparse
import logging

from menhir.tool import Tool
from menhir.tool_utils import (
    OK,
    FAIL,
    NOTHING_TO_DO,
    package_script,
    tool_env,
    working_dir,
)

log = logging.getLogger(__name__)


def tool():
    return Build()


class Build(Tool):

    def dir_info(tool, path, info):
        return {
            'project_recognised': True,
            'can_run': True,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build phase."""
        import subprocess

        log.info('build: %s %s %s', path, info, args)
        if (
                'changed' not in info or
                info['changed'].get('self') or
                info['changed'].get('dependents')
        ):
            log.info('Running build in %s', path)
            project_name = info['project-name']
            env = tool_env()
            env['MENHIR_PROJECT'] = project_name
            with package_script("/tools/build/test.sh") as f:
                with working_dir(path):
                    res = subprocess.call([f.name] + args.args, env=env)
                    if res:
                        return FAIL
                    return OK
        else:
            return NOTHING_TO_DO


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Invoke build based tests",
        **kwargs
    )
    parser.add_argument('args', nargs='*', help='build arguments')
    return parser



"""Tool that invokes the given command.

If the given command is given with a relative path, then this is
interpreted as a project local, and the failure is determined by the
``--if-no-op`` flag.

"""
import argparse
import logging
import subprocess

from os.path import exists, join

from menhir.tool import LINT, Tool
from menhir.tool_utils import (
    OK,
    FAIL,
    NOTHING_TO_DO,
    changed_state,
    has_self_or_dependent_changes,
    tool_env,
    working_dir
)

log = logging.getLogger(__name__)


def tool():
    return Pylama()


class Pylama(Tool):

    def dir_info(tool, path, info):
        has_setup_py = exists(join(path, 'setup.py'))
        return {
            'project_recognised': has_setup_py,
            'can_run': has_setup_py,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        """Arg parser for the tool options and arguments."""
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        """Pylamaute the script tool."""
        with working_dir(path):
            cmd = ['pylama']
            changed = changed_state(info)
            log.info('Running "%s" in %s', " ".join(cmd), path)
            if not has_self_or_dependent_changes(changed):
                return NOTHING_TO_DO

            env = tool_env()
            if changed is None:
                env['MENHIR_ALL'] = "1"
            else:
                if changed.get('self'):
                    env['MENHIR_CHANGED_SELF'] = "1"
                if changed.get('dependents'):
                    env['MENHIR_CHANGED_DEPENDENTS'] = "1"
                if changed.get('dependees'):
                    env['MENHIR_CHANGED_DEPENDEES'] = "1"

            result = subprocess.call(cmd, env=env)
            log.debug('Pylama result: %s', result)
            if result:
                return FAIL
            return OK

    def build_jobs(tool, info):
        return [LINT]


def parser(**kwargs):
    """Arg parser for the tool options and arguments."""
    parser = argparse.ArgumentParser(
        description="Run pylama.",
        **kwargs
    )

    parser.add_argument(
        'command_args',
        nargs=argparse.REMAINDER,
        help='command arguments'
    )
    return parser



"""The circleci tool writes a circle.yml file to build the project.

For multiproject repos:

    menhir multiproject --no-traverse circleci

Configuration is in `menhir-config.yaml`, under the `circleci` key.
Specify `jobs` and `build` as per CircleCI's `jobs` and `workspace.build` keys
for `config.yml`.


"""
import argparse
import logging
import os


import os.path

from collections import OrderedDict

from dialogue.multi_method import method, multi

from ..fileutils import yaml_dumper
from ..tool import (
    CHECKOUT,
    CHECKOUT_STATE,
    DEPENDENCIES,
    DEPLOY,
    LINT,
    TEST,
    Tool,
    build_impl,
    default_tools,
    load_tool,
)
from ..tool_utils import (
    OK,
    tool_env,
    working_dir,
)

log = logging.getLogger(__name__)

CIRCLECI = 'circleci'


def tool():
    return Circleci()


class Circleci(Tool):

    def dir_info(tool, path, info):
        log.debug('path: %s', path)
        log.info('info: %s', info)
        root = info.get('menhir', {}).get('root', '.')
        log.debug('path: %s, root %s', path, root)
        is_root = path == root
        return {
            'project_recognised': is_root,
            'can_run': is_root,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Write a config.yml file."""
        log.info('circleci: %s %s %s', path, info, args)
        log.info('Running circleci in %s', path)
        project_name = info['project-name']
        env = tool_env()
        env['MENHIR_PROJECT'] = project_name

        log.info('Info %s', info)

        build_info = build_info_for(
            info.get('circleci', {})
        )
        config = initial_config()

        if 'multiproject' in info:
            build_multi_config(
                config,
                build_info,
                info,
            )
        else:
            tools = tools_for(path, info)
            jobs = tool_jobs(tools, info)
            build_config(
                config,
                build_info,
                info,
                tools,
                jobs,
                default_build=DEFAULT_BUILD['single'],
            )

        yaml = yaml_dumper()

        with working_dir(path):
            config_dir = os.path.join(path, '.circleci')
            config_file = os.path.join(config_dir, 'config.yml')
            os.makedirs(config_dir, exist_ok=True)
            with open(config_file, 'w+') as f:
                # yaml.load(f)
                yaml.dump(config, f)
            return OK

    def traverse_multiproject(tool, info):
        return False


def tools_for(path, info):
    tools = info.get('menhir', {}).get('tools', default_tools())
    tool_impls = list(map(load_tool, tools))
    used_tools = list(filter(
        lambda t: t.dir_info(path, info).get('project_recognised'),
        tool_impls
    ))
    return used_tools


def tool_jobs(tools, info):
    jobs = []
    for tool in tools:
        jobs.extend(tool.build_jobs(info))
    return jobs


def build_info_for(info):
    log.debug('build_info_for  %s', info)
    build_info = {
        'default': docker(DEFAULT_DOCKER),
        'use_pyenv': False,
        'workspace': '/workspace',
    }
    build_info.update(info.get('build', {}))
    return build_info


def initial_config():
    config = OrderedDict(**PREAMBLE)
    config['jobs'] = {}
    config['workflows'] = {
        'version': 2,
        'build': {'jobs': []},
    }
    return config


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Write circleci config",
        **kwargs
    )
    parser.add_argument('args', nargs='*', help='circleci arguments')
    return parser


PREAMBLE = {
    'version': 2,
}

DEFAULT_DOCKER = {
    'image': 'hadrien/pythonci',
    'user': 'root',
}


def requires(jobs):
    return {
        'requires': jobs,
    }


DEFAULT_BUILD = {
    'single': {
        'jobs': [
            {CHECKOUT: {}},
            {DEPENDENCIES: requires([CHECKOUT])},
            {TEST: requires([DEPENDENCIES])},
            {LINT: requires([DEPENDENCIES])},
            {DEPLOY: requires([TEST, LINT])},
            # {'integration_test': requires(['deploy'])},
        ]
    },
    'multi': {
        'jobs': [
            {CHECKOUT: {}},
            {DEPENDENCIES: requires([CHECKOUT])},
            {TEST: requires([DEPENDENCIES])},
            {LINT: requires([DEPENDENCIES])},
            {DEPLOY: requires([TEST, LINT])},
            # {'integration_test': requires(['deploy'])},
        ]
    }
}

ROOT_TASKS = {CHECKOUT, CHECKOUT_STATE}


def build_multi_config(config, build_info, info):
    projects = info['multiproject']['project_meta']
    for path, meta in projects.items():
        deps = info['multiproject']['project_graph'][path]
        project_info = info['multiconfig']['project_infos'][path]
        tool_names = meta['tools']
        tools = list(map(load_tool, tool_names))
        jobs = tool_jobs(tools, info)
        build_config(
            config,
            build_info,
            project_info,
            tools,
            jobs,
            path=path,
            suffix=project_suffix(project_info['project-name']),
            deps=[
                project_suffix(dep)
                for dep in deps
            ],
            default_build=DEFAULT_BUILD['multi'],
        )


def project_suffix(project_name):
    return '_' + project_name


def build_config(
        config,
        build_info,
        info,
        tools,
        tool_jobs,
        path='.',
        suffix="",
        deps=[],
        default_build=DEFAULT_BUILD,
):
    """Update config with build tasks for project at path.

    path   - path to project, relative to project root
    suffix - project specific suffix for task names
    deps   - list of project paths of dependencies
    """
    log.debug('xx build_info %s', build_info)
    build_spec = build_info.get('build', default_build)
    build_jobs = []
    jobs = OrderedDict()

    for job in build_spec['jobs']:
        requires = []
        job_type, job_spec = list(job.items())[0]

        if job_type not in tool_jobs:
            continue

        if job_type in ROOT_TASKS:
            requires = job_spec.get('requires', [])
            job_name = job_type
        else:
            job_name = job_type + suffix
            for require in job_spec.get('requires', []):
                if require in ROOT_TASKS:
                    requires.append(require)
                else:
                    requires.append(require + suffix)
                    for dep in deps:
                        requires.append(require + dep)
        log.info('Write job: %s %s', job_name, job_spec)

        spec = {'requires': requires}
        build_flows, job_specs = spec_for(
            build_info, job_type, path, job_name, spec, info, tools,
        )
        log.info('flow, job: %s %s', build_flows, job_specs)
        build_jobs.extend(build_flows)
        jobs.update(job_specs)

    config['jobs'].update(jobs)
    config['workflows']['build']['jobs'].extend(build_jobs)

    return config


@multi
def spec_for(build_info, job_type, path, job_name, job_spec, info, tools):
    return job_type


def build_spec_for(
        build_info, job_type, path, job_name, job_spec, info, tools,
        load_cache_keys,
        save_cache_keys={},
        extra_steps=[],
):
    job_detail = OrderedDict(**build_info['default'])
    job_detail.update({
        'steps': [restore_cache(key) for key in load_cache_keys],
    })

    log.debug('xxinfo %s', info)
    steps = info_steps(info, job_type)
    log.debug('steps %s', steps)

    if steps:
        job_detail['steps'].extend(steps)
    else:
        invocations = [
            x
            for tool in tools
            for x in build_impl(tool, job_type, CIRCLECI, info)
        ]
        job_detail['steps'].extend(filter(lambda x: x, invocations))
        job_detail['steps'].extend(extra_steps)

    job_detail['steps'].extend(
        [
            save_cache(cache_key, paths)
            for cache_key, paths in save_cache_keys.items()
        ]
    )
    job_detail.update(working_directory(build_info, path))
    return [{job_name: job_spec}], {job_name: job_detail}


@method(spec_for)
def spec_for_default(
        build_info, job_type, path, job_name, job_spec, info, tools
):
    return build_spec_for(
        build_info, job_type, path, job_name, job_spec, info, tools,
        [source_cache_key(info), dependencies_cache_key(info)])


@method(spec_for, CHECKOUT)
def spec_for_checkout(
        build_info, job_type, path, job_name, job_spec, info, tools
):
    return build_spec_for(
        build_info, job_type, path, job_name, job_spec, info, tools,
        [], {source_cache_key(info): [build_info['workspace']]})


@method(spec_for, DEPENDENCIES)
def spec_for_dependencies(
        build_info, job_type, path, job_name, job_spec, info, tools
):
    if is_subproject(info):
        cache_paths = [
            "/home/circleci/.pyenv/versions/3.6.1/envs/"
            "{}/lib/python3.6/site-packages"
            .format(info["project-name"])
        ]
    else:
        cache_paths = [
            '/root/.cache/',
            '/usr/local/lib/python3.6/site-packages',
            '/usr/local/bin'
        ]
    return build_spec_for(
        build_info, job_type, path, job_name, job_spec, info, tools,
        [source_cache_key(info)],
        {dependencies_cache_key(info): cache_paths})


@method(spec_for, CHECKOUT_STATE)
def spec_for_checkout_state(
        build_info, job_type, path, job_name, job_spec, info, tools
):
    return build_spec_for(
        build_info, CHECKOUT, path, job_name, job_spec, info, tools,
        [], {source_cache_key(info): [build_info['workspace']]},
        extra_steps=[
            'menhir git_status multiproject write_state',
        ]
    )


def info_steps(info, job):
    return info.get('circleci', {}).get('jobs', {}).get(job, {}).get('steps')


def working_directory(build_info, path):
    return {
        'working_directory': os.path.join(build_info['workspace'], path),
    }


def docker(spec):
    return {
        'docker': [spec],
    }


def source_cache_key(info):
    return "repo-{{ .Environment.CIRCLE_SHA1 }}"


def dependencies_cache_key(info):
    name = info['project-name']
    return (
        'pydep-%s'
        # TODO - remove the SHA from the key
        '-{{ .Environment.CIRCLE_SHA1 }}'
        '-{{ checksum "requirements.txt" }}'
        '-{{ checksum "requirements-tests.txt" }}'
    ) % name


def restore_cache(cache_key):
    return {
        'restore_cache': {
            'key': cache_key,
        }
    }


def save_cache(cache_key, paths):
    return {
        'save_cache': {
            'key': cache_key,
            'paths': paths,
        }
    }


def run(name, invocations):
    return {
        'run': {
            'name': name,
            'command': "\n".join(invocations)
        },
    }


def is_subproject(info):
    return 'multiproject' in info


@method(build_impl, ('git', CHECKOUT, CIRCLECI))
def build_impl_git_checkout_circleci(tool, task, build_env, info):
    if is_subproject(info):
        return [
            {'checkout': {}},
            run('Write state', [
                'pip install menhir==$(cat .menhir-version)',
                'menhir git_status multiproject write_state',
            ]),

        ]
    else:
        return [{'checkout': {}}]


@method(build_impl, ('pyenv', DEPENDENCIES, CIRCLECI))
def build_impl_pyenv_dependencies_circleci(tool, task, build_env, info):
    is_multi = 'multiproject' in info
    if is_multi:
        return [run('Install Dependencies', ["menhir pyenv virtualenv"])]
    else:
        return [run(
            'Install Test Dependencies',
            ["[test -f requirements-tests.txt] && "
             "pip install -r requirements-tests.txt"])]


@method(build_impl, ('setup_py', DEPENDENCIES, CIRCLECI))
def build_impl_setup_py_dependencies_circleci(tool, task, build_env, info):
    is_multi = 'multiproject' in info
    if not is_multi:
        return [run('Install Dependencies', ["pip install ."])]
    else:
        return []


@method(build_impl, ('setup_py', TEST, CIRCLECI))
def build_impl_setup_py_test_circleci(tool, task, build_env, info):
    return [run('Test', ["menhir -v setup_py test"])]


@method(build_impl, ('codecov', TEST, CIRCLECI))
def build_impl_codecov_test_circleci(tool, task, build_env, info):
    return [run('Test', ["menhir -v codecov"])]


# @method(build_impl, ('pylama', LINT, CIRCLECI))
# def build_impl_pylama_lint_circleci(tool, task, build_env, info):
#     return [run('Test', ["menhir -v pylama"])]


@method(build_impl, ('docker', DEPLOY, CIRCLECI))
def build_impl_docker_push_circleci(tool, task, build_env, info):
    return [run('Deploy docker', [
        "$(aws --region ca-central-1 ecr get-login --no-include-email)",
        "menhir docker publish",
    ])]



"""Encrypt and decrypt data using KMS keys.

Reads configuration from the ``kms`` key.

Configuration
-------------

The ``targets`` key specifies a map for each target.

Each target can have the following keys:

:``file``:        The file to encypt to, or decrypt from.
:``key``:         The KMS key name to use.
:``data-key``:    The data-key to decrypt to.
:``format``:      The format to parse the data as for data-key.
:``plain-file``:  The file to decrypt to, or encrypt from.
:``encode``:      Encoding for encrypted file (none, or base64).
:``region_name``: AWS region for KMS key.

:``context``: A list of configuration names that will be used to generate an
              encryption context.  The default is ``[project]``.

:``envelope``: The file to write the encypted key to.  If no envelope is
               specified, then the data is encrypted with the KMS key
               directly (4Kb limit).


Tasks
-----

:``encrypt``:      is used to encrypt a file.
:``decrypt``:      is used to decrypt a file.
:``decrypt-data``: is used to decrypt a file to a ``data`` key in menhir.

"""
import argparse
import json
import logging
import subprocess

from base64 import b64encode, b64decode
from os.path import exists, join

import boto3
from ruamel.yaml import YAML

from menhir.fileutils import load_encrypted, slurp_bytes, spit_bytes
from menhir.tool import Tool
from menhir.tool_config import aliased_value_array
from menhir.utils import getd, multi, method, select_keys
from menhir.tool_utils import (
    OK,
    FAIL,
    NOTHING_TO_DO,
)

log = logging.getLogger(__name__)


def tool():
    return Kms()


class Kms(Tool):

    def dir_info(tool, path, info):
        return {
            'project_recognised': False,
            'can_run': True,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        return parser(**kwargs)

    def execute_tool(tool, path, info, args,):
        """Execute a build task."""
        return task(path, info, args)


def parser(**kwargs):
    parser = argparse.ArgumentParser(
        description="Commands to encrypt and decrypt data using KMS",
        **kwargs
    )
    parsers = parser.add_subparsers(help="Kms commands", dest='task')

    p = parsers.add_parser(
        'encrypt-file',
        help='Encrypt a file to using a KMS key.'
    )
    p.add_argument('-p', '--plain-file', help="Source file to encrypt")

    g = p.add_mutually_exclusive_group()
    g.add_argument('-f', '--file', help="File name for the encrypted data")
    g.add_argument('-t', '--target', help="Target name")

    p.add_argument('-k', '--key', help="KMS key name")
    p.add_argument('-e', '--envelope', help="File name for the encrypted key")
    p.add_argument(
        '-c', '--context',
        help="Specify a list of config keys for the encryption context"
    )
    p.add_argument(
        '-n', '--encode',
        choices=['none', 'base64'],
        help="Encode encrypted file"
    )
    p.add_argument(
        '--arg', dest='args', action='append',
        help="Specify an argument for the encryption settings")
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    p = parsers.add_parser(
        'decrypt-file',
        help='Decrypt a file to using a KMS key.'
    )

    g = p.add_mutually_exclusive_group(required=True)
    g.add_argument('-t', '--target', help="Target name")
    g.add_argument('-f', '--file', help="Encrypted data file name")

    g = p.add_mutually_exclusive_group()
    g.add_argument(
        '-p', '--plain-file',
        help="File to write the decrypted data to"
    )
    g.add_argument('-d', '--data-key', help="Menhir data key to write to")
    p.add_argument(
        '-m', '--format',
        help="Format to parse data written to key"
    )

    p.add_argument('-k', '--key', help="KMS key name")
    p.add_argument('-e', '--envelope', help="File name for the encrypted key")
    p.add_argument(
        '-c', '--context',
        help="Specify a list of config keys for the encryption context"
    )
    p.add_argument(
        '-n', '--encode',
        choices=['none', 'base64'],
        help="Encrypted file is encoded"
    )
    p.add_argument(
        '--arg', dest='args', action='append',
        help="Specify an argument for the decryption settings")
    p.add_argument('remainder', nargs=argparse.REMAINDER)

    return parser


@multi
def task(path, info, args):
    log.debug("kms task: %s", args.task)
    return args.task


@method(task, 'encrypt-file')
def task_encrypt(path, info, args):
    if not (args.target or args.file):
        log.error('No target or file specified')
        return FAIL

    config = info.get('kms', {}).get('targets', {}).get(args.target, {})

    arg_dict, status = target_args(config, args)
    log.debug("arg_dict %s", arg_dict)
    if status != OK:
        return status

    encrypt_config_for_args(config, args)

    key = config.get('key')
    if not key:
        log.error(
            'No KMS encryption key name specified for target %s',
            config.get('file')
        )
        return FAIL

    values_list = aliased_value_array(
        config.get('context', []),
        info,
        path,
        arg_dict,
    )
    context = {k: v for k, v in values_list}

    arg_dict.update(context)
    expand_keys(config, arg_dict)

    log.debug('encrypt: %s', config)
    if config.get('envelope'):
        return encrypt_with_envelope(config, context)
    else:
        return encrypt_with_master(config, context)


def encrypt_config_for_args(config, args):
    if args.file:
        config['file'] = args.file
    if args.encode:
        config['encode'] = args.encode
    if args.key:
        config['key'] = args.key
    if args.plain_file:
        config['plain-file'] = args.plain_file
    if args.envelope:
        config['envelope'] = args.envelope
    if args.context:
        config['context'] = list(map(
            lambda x: x.strip(),
            args.context.split(",")
        ))


def encrypt_with_master(config, context):
    data = slurp_bytes(config['plain-file'])
    client_config = select_keys(config, ['region_name'])
    kms = boto3.client('kms', **client_config)
    res = kms.encrypt(
        KeyId=config['key'],
        Plaintext=data,
        EncryptionContext=context,
    )
    encode = config.get('encode', 'base64')
    if encode == 'none':
        data = res['CiphertextBlob']
    elif encode == 'base64':
        data = b64encode(res['CiphertextBlob'])

    spit_bytes(config['file'], data)
    return OK


def encrypt_with_envelope(config, context):
    client_config = select_keys(config, ['region_name'])
    kms = boto3.client('kms', **client_config)
    envelope_path = config['envelope']
    if exists(envelope_path):
        data = slurp_bytes(envelope_path)
        data = b64decode(data)
        key = kms.decrypt(
            CiphertextBlob=data,
            EncryptionContext=context,
        )
        plain_key = key['Plaintext']
    else:
        key = kms.generate_data_key(
            KeyId=config['key'],
            KeySpec='AES_256',
            EncryptionContext=context,
        )
        plain_key = key['Plaintext']
        enc_key = b64encode(key['CiphertextBlob'])
        spit_bytes(envelope_path, enc_key)

    return local_encrypt(
        plain_key,
        config['plain-file'],
        config['file'])


def local_encrypt(key, source, dest):
    res = subprocess.call([
        "openssl", "aes-256-cbc", "-a",
        "-k", key,
        "-in", source,
        "-out", dest
    ])
    if res == 0:
        return OK
    return FAIL


@method(task, 'decrypt-file')
def task_decrypt(path, info, args):
    if not (args.target or args.file):
        log.error('No target or encrypted file specified')
        return FAIL

    config = info.get('kms', {}).get('targets', {}).get(args.target, {})
    if args.target and not config:
        log.debug("kms decrypt: nothing to do for target %s", args.target)
        return NOTHING_TO_DO

    arg_dict, status = target_args(config, args)
    log.debug("arg_dict %s", arg_dict)
    if status != OK:
        return status

    decrypt_config_for_args(config, args)

    key = config.get('key')
    if not key:
        log.error(
            'No KMS decryption key name specified for target file %s',
            config['file']
        )
        return FAIL

    values_list = aliased_value_array(
        config.get('context', []),
        info, path, arg_dict
    )
    context = {k: v for k, v in values_list}

    arg_dict.update(context)
    expand_keys(config, arg_dict)

    log.debug('decrypt: %s', config)
    if config.get('envelope'):
        return decrypt_with_envelope(path, info, config, context)
    else:
        return decrypt_with_master(path, info, config, context)


def decrypt_config_for_args(config, args):
    if args.file:
        config['file'] = args.file
    if args.plain_file:
        config['plain_file'] = args.plain_file

    if args.key:
        config['key'] = args.key
    if args.data_key:
        config['data-key'] = args.data_key
    if args.format:
        config['format'] = args.format
    if args.envelope:
        config['envelope'] = args.envelope
    if args.encode:
        config['encode'] = args.key
    if args.context:
        config['context'] = list(map(
            lambda x: x.strip(),
            args.context.split(",")
        ))


def decrypt_with_master(path, info, config, context):
    data = slurp_bytes(join(path, config['file']))
    if config.get('encode', 'base64') == 'base64':
        data = b64decode(data)
    client_config = select_keys(config, ['region_name'])
    kms = boto3.client('kms', **client_config)
    res = kms.decrypt(
        CiphertextBlob=data,
        EncryptionContext=context,
    )
    data_key = config.get('data-key')
    if data_key and not config.get('plain_file'):
        data = res['Plaintext']
        log.debug('data is %s', data)
        parse_as = config.get('format', 'json')
        if parse_as == 'json':
            data = json.loads(data)
        elif parse_as == 'yaml':
            yaml = YAML()
            data = yaml.load(data)
        else:
            log.error('Unknown foramt %s', parse_as)
            return FAIL
        getd(info, 'data')[data_key] = data
    else:
        spit_bytes(join(path, config['plain_file']), res['Plaintext'])

    return OK


def decrypt_with_envelope(path, info, config, context):
    client_config = select_keys(config, ['region_name'])
    kms = boto3.client('kms', **client_config)
    enc_key = b64decode(slurp_bytes(config['envelope']))
    key = kms.decrypt(
        CiphertextBlob=enc_key,
        EncryptionContext=context,
    )
    data_key = config.get('data-key')

    if data_key and not config.get('plain_file'):
        data = load_encrypted(join(path, config['file']), key['Plaintext'])
        parse_as = config.get('format', 'json')
        if parse_as == 'json':
            data = json.loads(data)
        elif parse_as == 'yaml':
            yaml = YAML()
            data = yaml.load(data)
        else:
            log.error('Unknown foramt %s', parse_as)
            return FAIL
        getd(info, 'data')[data_key] = data
        return OK
    else:
        return local_decrypt(
            key['Plaintext'],
            join(path, config['file']),
            join(path, config['plain_file'])
        )


def local_decrypt(key, source, dest):
    res = subprocess.call([
        "openssl", "aes-256-cbc", "-a", "-d",
        "-k", key,
        "-in", source,
        "-out", dest
    ])
    if res == 0:
        return OK
    return FAIL


def target_args(config, args):
    """Return the target arguments as a dict."""
    expected_args = config.get('args', [])
    actual_args = args.args or []
    if len(expected_args) != len(actual_args):
        log.error('Exepcted decrypt --arg for each of %s', expected_args)
        return None, FAIL
    else:
        return {k: v for (k, v) in zip(expected_args, actual_args)}, OK


def expand_keys(config, arg_dict):
    for k in ['file', 'key', 'plain_file', 'envelope', 'region_name']:
        v = config.get(k)
        if v:
            config[k] = v % arg_dict



"""Utility functions."""

# This module should have no dependencies on other menhir modules

# Multi-dispatch functions for python
# Modified from https://gist.github.com/adambard/6bf15282874ca89404f7


def multi(dispatch_fn, default=None):
    """Decorate a function to dispatch on.

    The value returned by the dispatch function is used to look up the
    implementation function based on its dispatch key.
    """
    def _inner(*args, **kwargs):
        dispatch_value = dispatch_fn(*args, **kwargs)
        f = _inner.__multi__.get(dispatch_value, _inner.__multi_default__)
        if f is None:
            raise Exception(
                "No implementation of %s for dispatch value %s" % (
                    dispatch_fn.__name__,
                    dispatch_value
                )
            )
        return f(*args, **kwargs)

    _inner.__multi__ = {}
    _inner.__multi_default__ = default
    return _inner


def method(dispatch_fn, dispatch_key=None):
    """Decorate a function implementing dispatch_fn for dispatch_key.

    If no dispatch_key is specified, the function is used as the
    default dispacth function.
    """
    def apply_decorator(fn):
        if dispatch_key is None:
            # Default case
            dispatch_fn.__multi_default__ = fn
        else:
            dispatch_fn.__multi__[dispatch_key] = fn
        return dispatch_fn
    return apply_decorator


def reductions(f, lst, init):
    """Like reduce, but returns all intermediate values as a sequence."""
    prev = init
    for i in lst:
        prev = f(prev, i)
        yield prev


# http://stackoverflow.com/questions/7204805/python-dictionaries-of-dictionaries-merge
# merges b into a

def deep_merge(a, b, path=None):
    """Deep merge b into a.

    Modifies a.  Returns the modified dict a.
    """
    if path is None:
        path = []
    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                deep_merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            elif isinstance(a[key], list) and isinstance(b[key], list):
                a[key].extend(b[key])
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


def getd(d, k):
    """Get a dict from key, k, of dict, d, creating a dict if needed.

    This function replaces the `v = d.get(k,{}); d[k] = v` pattern.
    """
    value = d.get(k)
    if value is None:
        value = d[k] = {}
    return value


def select_keys(d, ks):
    """Return a dict containing just the specified keys, ks, from d."""
    return {k: d[k] for k in ks if k in d}


try:
    _basestring = basestring
except NameError:
    _basestring = str


def is_string(x):
    """Predicate for value being a string."""
    return isinstance(x, _basestring)



"""Add git status information.

The :command:`git_status` tool adds a list of changed files into the
menhir state dict.  This is used by th command:`multiproject` tool to
determine which projects, dependents and dependees have been changed.

"""
import argparse
import logging

from menhir import gitutils
from menhir.tool import Tool
from menhir.tool_utils import OK

log = logging.getLogger(__name__)


def tool():
    return GitStatus()


class GitStatus(Tool):

    def name(tool):
        return 'git_status'

    def dir_info(tool, path, info):
        from menhir import gitutils
        has_git = gitutils.find_root()
        return {
            'project_recognised': has_git,
            'can_run': has_git,
        }

    def dependencies(tool, path):
        return []

    def arg_parser(tool, **kwargs):
        """Arg parser for the tool options and arguments."""
        return parser(**kwargs)

    def execute_tool(tool, path, info, args):
        """Add git status information."""
        git_root = gitutils.find_root()
        files = changed_files(git_root, args.from_commit)
        info['git_status'] = files
        info['git_status']['root'] = git_root
        return OK


def parser(**kwargs):
    """Arg parser for the tool options and arguments."""
    parser = argparse.ArgumentParser(
        description="""GitStatus command to add git status.
Adds changed information to project infos.
        """,
        **kwargs
    )
    parser.add_argument(
        "--from-commit",
        help="""
Run task on projects that depend on files changed since the
specified commit.  This can be specified as a single commit,
e.g. 3f8e5bb, 3f8e5bb^, head^.
"""
    )

    parser.add_argument(
        "--dirty",
        help="""
Run task on projects that depend on dirty (uncommited) files.
"""
    )
    parser.add_argument('remainder', nargs=argparse.REMAINDER)
    return parser


def changed_files(git_root, from_commit):
    """Return the files that have been changed.

    If from_commit is not None, or an implicit from_commit is found
    from CIRCLE_COMPARE_URL, or CIRCLE_SHA1, then return files changed
    since that commit.

    If there are uncommitted files, return them, return the last
    commit's files.
    """
    from menhir.gitutils import (
        files_changed_since,
        head_commit,
        repo,
        uncommited_files,
    )
    git_repo = repo(git_root)

    from_commit = with_implicit_from_commit(from_commit) or \
        str(head_commit(git_repo)) + "^"

    uncommited = uncommited_files(git_repo)['all']

    files = files_changed_since(git_repo, from_commit)

    return {
        'from_commit': from_commit,
        'changed': files,
        'uncommited': uncommited,
    }


def with_implicit_from_commit(from_commit):
    """Return any implicit from commit specification.

    Returns the first of the following that is not nil:

    * ``from_commit``
    * the start of the CIRCLE_COMPARE_URL range
    * the commit prior to CIRCLE_SHA1
    """
    import os

    if from_commit:
        return from_commit

    compare_url = os.getenv('CIRCLE_COMPARE_URL')
    if compare_url:
        log.debug('Deriving commit range from %s', compare_url)
        commit_range = compare_url.rsplit('/', 1)[1]
        return commit_range.split('..')[0]

    if os.getenv('CIRCLE_SHA1'):
        sha = os.getenv('CIRCLE_SHA1')
        return sha+"^"


# def git_config():
#     from os.path import relpath
#     from menhir import gitutils
#     git_root = gitutils.find_root()
#     return {
#         'git_root': git_root,
#         'git_prefix': relpath(git_root),
#     }
